/*
 * USB Orinoco driver
 *
 * Copyright (c) 2003 Manuel Estrada Sainz <ranty@debian.org>
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License
 * at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and
 * limitations under the License.
 *
 * Alternatively, the contents of this file may be used under the
 * terms of the GNU General Public License version 2 (the "GPL"), in
 * which case the provisions of the GPL are applicable instead of the
 * above.  If you wish to allow the use of your version of this file
 * only under the terms of the GPL and not to allow others to use your
 * version of this file under the MPL, indicate your decision by
 * deleting the provisions above and replace them with the notice and
 * other provisions required by the GPL.  If you do not delete the
 * provisions above, a recipient may use your version of this file
 * under either the MPL or the GPL.  
 *
 * Initialy based on USB Skeleton driver - 0.7
 *
 * Copyright (c) 2001 Greg Kroah-Hartman (greg@kroah.com)
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation; either version 2 of
 *	the License, or (at your option) any later version.
 *
 * NOTE: The original USB Skeleton driver is GPL, and my work is
 * MPL/GPL which efectively means that all of it can just be used
 * under the terms of the GPL. All contributions will fall on the
 * MPL/GPL by default. Once all the GPL only code is gone it can all
 * be release as MPL/GPL.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/poll.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/fcntl.h>
#include <linux/module.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/smp_lock.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/usb.h>
#include <linux/timer.h>
#include <asm/io.h>

#include <linux/netdevice.h>
#include <linux/if_arp.h>
#include <linux/etherdevice.h>
#include <linux/wireless.h>

#include "hermes.h"
#include "hermes_rid.h"
#include "orinoco.h"

#include "fwfs.h"
#include "orinoco_usb.h"
#include "orinoco_usb_fw.h"

#ifdef CONFIG_USB_DEBUG
	static int debug = 1;
#else
	static int debug;
#endif

/* Use our own dbg macro */
#undef dbg
#define dbg(format, arg...) do { if (debug) printk(KERN_DEBUG __FILE__":"__FUNCTION__ ": " format "\n" , ## arg); } while (0)

#define ndbg(format, arg...)
#define dump(a, b) usb_skel_debug_data(__FUNCTION__, b, a)
#define ndump(a, b)

#define UNUSED __attribute__((unused))
#ifndef MIN
#define MIN(a, b) (((a) < (b)) ? (a) : (b))
#endif
#ifndef MAX
#define MAX(a, b) (((a) > (b)) ? (a) : (b))
#endif

/* Version Information */
#define DRIVER_VERSION "v0.0.7"
#define DRIVER_DESC "Orinoco USB Bridge Driver"

/* Module paramaters */
MODULE_PARM(debug, "i");
MODULE_PARM_DESC(debug, "Debug enabled or not");

/* Some times, the card gets stuck and stops paying
 * attention to the world until we do something to it. Setting
 * watchdog_secs to a nonzero value will set a timer on the USB
 * callbacks that will trigger within watchdog_secs seconds of
 * inactivity. The method to get the card back to work is just to send an
 * INQ_TALLIES command which should be harmless */
static int watchdog_secs = 0;
MODULE_PARM(watchdog_secs, "i");
MODULE_PARM_DESC(watchdog_secs, "0 to disable");

static int force_unsupported = 0;
MODULE_PARM(force_unsupported, "i");
MODULE_PARM_DESC(force_unsupported, "Try to handle untested hardware.");

static int default_card_variant = 0;
MODULE_PARM(default_card_variant, "i");
MODULE_PARM_DESC(default_card_variant, "Card variant to use as a first guess");
/* Most of this Device/Vendor pairs are untested */

#define USB_COMPAQ_VENDOR_ID		0x049f /* Compaq Computer Corp. */
#define USB_COMPAQ_WL215_ID		0x001f /* Compaq WL215 USB Adapter */
#define USB_COMPAQ_W200_ID		0x0076 /* Compaq W200 USB Adapter */

#define USB_BUFFALO_VENDOR_ID		0x0411
#define USB_BUFFALO_AIRSTA_ID		0x000B /* Buffalo airstation usb */

#define USB_LUCENT_VENDOR_ID		0x047E /* Lucent Technologies */
#define USB_LUCENT_ORINOCO_ID		0x0300 /* Lucent/Agere Orinoco USB
						  Client */
#define USB_AVAYA8_VENDOR_ID		0x0D98
#define USB_AVAYAE_VENDOR_ID		0x0D98
#define USB_AVAYA_WIRELESS_ID		0x0300 /* Avaya Wireless USB Card */

#define USB_AGERE_VENDOR_ID		0x0D4E /* Agere Systems */
#define USB_AGERE_ORINOCO_ID		0x1000

#define USB_ELSA_VERNDOR_ID		0x05CC
#define USB_ELSA_AIRLANCER_ID		0x3100 /* ELSA AirLancer USB-11 */

#define USB_LEGEND_VENDOR_ID		0x0E7C
#define USB_LEGEND_JOYNET_ID		0x0300 /* LEGEND Joynet WLAN USB Card */

#define USB_SAMSUNG_VENDOR_ID		0x04E8
#define USB_SAMSUNG_SEW2001U_ID		0x5002 /* Samsung SEW-2001u Card */
#define USB_SANSUNG_SEW2001U2_ID	0x5B11 /* Samsung SEW-2001u Card */

#define USB_IGATE_VENDOR_ID		0x0681
#define USB_IGATE_IGATE_11M_ID		0x0012 /* I-GATE 11M USB Card */

#define USB_FUJITSU_VENDOR_ID		0x0BF
#define USB_FUJITSU_E1100_ID		0x1002 /* Fujitsu Siemens Computers
						  connect2AIR WLAN E-1100 USB */
#define USB_2WIRE_VENDOR_ID		0x1630
#define USB_2WIRE_WIRELESS_ID		0xff81 /* 2Wire Wireless USB adapter */


/* table of devices that may work with this driver */
/* We will refuse to touch any untested device unless force_unsupported is set
 */
static struct usb_device_id bridge_table [] = {
	{ USB_DEVICE(USB_COMPAQ_VENDOR_ID, USB_COMPAQ_WL215_ID) },
	{ USB_DEVICE(USB_COMPAQ_VENDOR_ID, USB_COMPAQ_W200_ID) },
	{ USB_DEVICE(USB_BUFFALO_VENDOR_ID, USB_BUFFALO_AIRSTA_ID) },
	{ USB_DEVICE(USB_LUCENT_VENDOR_ID, USB_LUCENT_ORINOCO_ID) },
	{ USB_DEVICE(USB_AVAYA8_VENDOR_ID, USB_AVAYA_WIRELESS_ID) },
	{ USB_DEVICE(USB_AVAYAE_VENDOR_ID, USB_AVAYA_WIRELESS_ID) },
	{ USB_DEVICE(USB_AGERE_VENDOR_ID, USB_AGERE_ORINOCO_ID) },
	{ USB_DEVICE(USB_ELSA_VERNDOR_ID, USB_ELSA_AIRLANCER_ID) },
	{ USB_DEVICE(USB_LEGEND_VENDOR_ID, USB_LEGEND_JOYNET_ID) },
	{ USB_DEVICE_VER(USB_SAMSUNG_VENDOR_ID, USB_SAMSUNG_SEW2001U_ID,
			 0, 0) },
	{ USB_DEVICE(USB_SAMSUNG_VENDOR_ID, USB_SANSUNG_SEW2001U2_ID) },
	{ USB_DEVICE(USB_IGATE_VENDOR_ID, USB_IGATE_IGATE_11M_ID) },
	{ USB_DEVICE(USB_FUJITSU_VENDOR_ID, USB_FUJITSU_E1100_ID) },
	{ USB_DEVICE(USB_2WIRE_VENDOR_ID, USB_2WIRE_WIRELESS_ID) },
	{ USB_DEVICE(0x0D4E, 0x047A) },
	{ }					/* Terminating entry */
};

MODULE_DEVICE_TABLE (usb, bridge_table);

/* Devices should be added to 'bridge_table' and one of
 * bridge_varian1_table or bridge_varian2_table. Devices on the
 * 'bridge_table' but not on any of 'bridge_varian1_table' or
 * 'bridge_varian2_table' will require force_unsupported=1.  */

/* Variant 1 has a PCMCIA card inside */
static struct usb_device_id bridge_varian1_table [] = {
	{ USB_DEVICE(USB_COMPAQ_VENDOR_ID, USB_COMPAQ_WL215_ID) },
	{ USB_DEVICE(USB_COMPAQ_VENDOR_ID, USB_COMPAQ_W200_ID) },
	{ }
};
/* Variant 2 is all build in a single board */
static struct usb_device_id bridge_varian2_table [] = {
	{ }
};

struct bridge_read {
	/* the RID that we are waiting for */
	u16 wait_rid;
	int done;
	struct urb *urb;
	//int urb_busy;
	u8 pipe;
	int buff_size;
	unsigned char *buff[2];
	int avail_buff;
	int avail_actual_length;
	int avail_status;
	struct {
		int offset;
	} bap;
};
struct bridge_write {
	unsigned char *buff;
	int buff_size;
	struct urb *urb;
	u8 pipe;
	int tx_size;
};
enum bridge_priv_flags {
	ORINOCO_USB_NEEDS_AC_REQ=1,
};	
/* Structure to hold all of our device specific stuff */
struct bridge_priv {
	struct usb_device *udev;
	struct usb_interface *interface;
	struct net_device *net_dev;
	struct bridge_read read;
	struct bridge_write write;
	u8 reply_count;
	struct timer_list watchdog_timer;
	struct semaphore sem;
	u8 card_variant;
	u16 hermes_reg_fake [0x40];
	int (*orinoco_stop)(struct net_device *dev);
	int (*orinoco_open)(struct net_device *dev);
	u32 flags;
};

/* the global usb devfs handle */
extern devfs_handle_t usb_devfs_handle;

/* Holder for the firmwarefs data entry */
static const struct fwfs_entry *fwfs_entry;

/* local function prototypes */
static void * skel_probe	(struct usb_device *dev, unsigned int ifnum,
				const struct usb_device_id *id);
static void skel_disconnect	(struct usb_device *dev, void *ptr);

static void bridge_write_bulk_callback	(struct urb *urb);
static void bridge_read_bulk_callback (struct urb *urb);

static inline void usb_skel_debug_data (const char *function, int size,
					const unsigned char *data);

static char *rid_name(u16 rid);

/* lock to protect the minor_table structure */
static DECLARE_MUTEX (minor_table_mutex);
/* usb specific object needed to register this driver with the usb subsystem */
static struct usb_driver skel_driver = {
	name:		"Orinoco USB",
	probe:		skel_probe,
	disconnect:	skel_disconnect,
	id_table:	bridge_table,
};

/*
 * Orinoco helper functions
 */

static __inline__ u16 build_crc(u8 *data)
{
	u16 crc = 0;
	int i;
	for (i=0; i<8; i++){
		crc = (crc<<1) + data[i];
	}
	return crc;
}

static __inline__ u8 bridge_reply_inc(u8 count)
{
	if (count < 0x7F)
		return count+1;
	else
		return 1;
}

/**
 * fill_bridge_req:
 * 
 * if data == NULL and length > 0 the data is assumed to be already in
 * the target buffer and only the header is filled.
 *
 */
static int fill_bridge_req(u8 *buf, u16 length, u16 rid, const void *data,
			   u16 x04, u8 reply_count)
{
	struct bridge_req_packet *req;
	int total_size = sizeof(*req)+length;

	if(total_size > BULK_BUF_SIZE){
		err("bulk request too big");
		return -EINVAL;
	}

	req = (struct bridge_req_packet *)buf;
	req->x00 = 0x0210;
	req->reply_count = reply_count;
	req->x03 = 0x00;
	req->x04 = x04;
	req->x05 = 0x00;
	req->size= length + 4;
	req->crc = build_crc((u8*)req);
	req->hermes_len = (length+3)/2;
	req->hermes_rid = rid;
	if(data)
		memcpy(req->data, data, length);
	return total_size;
}

static int bridge_submit_in_urb(struct bridge_priv *dev)
{
	int retval = 0;
	ndbg("ENTER");
	if(dev->read.urb->status == -EINPROGRESS){
		ndbg("urb busy, not resubmiting");
		retval = -EBUSY;
		goto exit;
	}
	dev->read.avail_actual_length = dev->read.urb->actual_length;
	dev->read.avail_status = dev->read.urb->status;
	dev->read.avail_buff = !dev->read.avail_buff;
	usb_fill_bulk_urb(dev->read.urb, dev->udev,
			  usb_rcvbulkpipe(dev->udev, dev->read.pipe),
			  dev->read.buff[!dev->read.avail_buff],
			  dev->read.buff_size,
			  bridge_read_bulk_callback, dev);
	dev->read.urb->timeout = 0; //300*HZ;
	dev->read.urb->transfer_flags = 0; //USB_ASYNC_UNLINK;
        retval = usb_submit_urb(dev->read.urb);
        if (retval) {
		err(__FUNCTION__ "submit failed %d", retval);
		goto exit;
        }
exit:
	ndbg("EXIT");
	return retval;
}
static int bridge_remove_in_urb(struct bridge_priv *dev)
{
	int retval = 0;
	ndbg("ENTER");
	if (dev->read.urb->status != -EINPROGRESS){
		dbg("no urb to remove");
		retval = -ENOENT;
		goto exit;
	}
	retval = usb_unlink_urb(dev->read.urb);
exit:	
	ndbg("EXIT");
	return retval;
}
static __inline__ int ezusb_8051_cpucs(struct bridge_priv *dev, u8 reset)
{
	if(!dev->udev){
		err(__FUNCTION__ ": !dev->udev");
		return -EFAULT;
	}
	return usb_control_msg(dev->udev,
			       usb_sndctrlpipe(dev->udev, 0),
			       EZUSB_REQUEST_FW_TRANS,
			       USB_TYPE_VENDOR|USB_RECIP_DEVICE|USB_DIR_OUT,
			       EZUSB_CPUCS_REG, 0x0,
			       &reset, sizeof(reset),
			       DEF_TIMEOUT);
}

static int bridge_reset(struct bridge_priv *dev)
{
	int retval = 0;
	dbg("ENTER");
	retval = ezusb_8051_cpucs(dev, 1);
	if (retval < 0){
		err("Failed to reset");
		goto exit;
	}

	retval = ezusb_8051_cpucs(dev, 0);
	if (retval < 0){
		err("Failed to unreset");
		goto exit;
	}
exit:
	dbg("EXIT");
	return retval;
}

static int find_fw_variant_offset(struct ez_usb_fw *fw)
{
	int offset;
	offset = *(short *)(fw->code+1);
	offset = be16_to_cpu(offset);
	for(; offset < fw->size; offset++){
		if (memcmp("\x02\x03\x58", &fw->code[offset], 3)==0)
			break;
	}
	if (offset > (fw->size-4)){
		dbg("couldn't find card_variant_offset");
		return -1;
	}
	fw->card_variant_offset = offset + 3;
	dbg("card_variant_offset = 0x%04X", fw->card_variant_offset);
	return 0;
}

static int bridge_firmware_download(struct bridge_priv *dev,
				    struct ez_usb_fw *fw)
{
	int retval, addr;
	int packet_size = dev->udev->descriptor.bMaxPacketSize0;
	int variant_offset = fw->card_variant_offset%packet_size;
	int variant_base = fw->card_variant_offset - variant_offset;
        u8 *transfer_buffer =  kmalloc (packet_size, GFP_KERNEL);

	dbg("ENTER");
	dbg("packet size: %d", packet_size);
        if (!transfer_buffer) {
                err("orinoco_usb: kmalloc(%d) failed.", packet_size);
                retval = -ENOMEM;
		goto fail;
        }

	retval = ezusb_8051_cpucs(dev, 1);
	if(retval < 0)
		goto fail;
	for(addr = 0; addr < fw->size; addr += packet_size){
		if((addr >= 0x100) && (addr < 0x300)){
			/* 0x100-0x300 should be left alone, it
			 * contains card specific data, like USB
			 * enumeration information */
			ndbg("Skiping 0x100-0x300 (0x%04X)", fw_rec->Address);
			continue;
		}
		memcpy (transfer_buffer, &fw->code[addr], packet_size);
		if(addr == variant_base){
			dbg("Patching card_variant byte 0x%02X at 0x%04X",
			    dev->card_variant, fw->card_variant_offset);
			transfer_buffer[variant_offset] = dev->card_variant;
		}
		retval = usb_control_msg(dev->udev,
					 usb_sndctrlpipe(dev->udev, 0),
					 EZUSB_REQUEST_FW_TRANS,
					 USB_TYPE_VENDOR|USB_RECIP_DEVICE
					 |USB_DIR_OUT,
					 addr, 0x0,
					 transfer_buffer, packet_size,
					 DEF_TIMEOUT);

		if(retval < 0)
			goto fail;
	}
	retval = ezusb_8051_cpucs(dev, 0);
	goto exit;
fail:
	err("orinoco_usb: fw upload: retval=%d", retval);
exit:
	if(transfer_buffer)
		kfree (transfer_buffer);
	dbg("EXIT");
	return retval;
}
static int bridge_access_ltv(struct bridge_priv *dev, u16 length, u16 rid,
			     const void *data, u16 x04,
			     u16 ans_rid, void *ans_buff, int ans_size,
			     u16 *ans_length, const int locked)
{
	struct net_device *net_dev = dev->net_dev;
	struct orinoco_private *priv = net_dev->priv;
	hermes_t *hw = &priv->hw;
	struct bridge_ans_packet *ans;
	int req_size;
	int retval = 0;
	int timeout = DEF_TIMEOUT;
	unsigned long flags=0;

	if(!dev->udev){
		dbg("Device disconnected");
		return -ENODEV;
	}

	if (!locked)
		spin_lock_irqsave(&hw->lock, flags);

	if(dev->read.urb->status != -EINPROGRESS){
		err(__FUNCTION__ ": in urb not pending");
	}

	req_size = fill_bridge_req(dev->write.buff, length, rid, data,
				   x04, dev->reply_count);
	if (req_size < 0){
		retval = req_size;
		goto exit;
	}
	ndump(dev->write.buff, length+sizeof(struct bridge_req_packet));

        dev->read.done = 0;
	dev->read.wait_rid = ans_rid;

	usb_fill_bulk_urb(dev->write.urb, dev->udev, 
			  usb_sndbulkpipe(dev->udev, dev->write.pipe),
			  dev->write.buff, req_size,
			  bridge_write_bulk_callback, dev);
	retval = usb_submit_urb(dev->write.urb);
	if (ans_rid == 0){
		/* there should be no answer */
		ndbg("Not waiting for an answer");
		goto exit;
	}
	if (retval){
		err("BULK OUT retval: %d", retval);
		goto exit;
	}

	if (!locked)
		spin_unlock_irqrestore(&hw->lock, flags);

        while (timeout && !dev->read.done)
        {
		/* We will mostly be called from bh or bh atomic
		 * context we can't sleep anyway, so we just busy wait
		 * for the interrupt. And for network trafic, we don't
		 * even get here */
		/* sleep for 5 jiffies each time */
		/* NOTE: will not be precise with big or small
		 * HZ, but it should be ok */
		mdelay(5000/HZ);
		timeout -= HZ/5;
		if(timeout != (DEF_TIMEOUT-HZ/5)){
			dbg("timeout %d flags: 0x%lX", timeout, flags);
			bridge_submit_in_urb(dev);
		}
        }

	if (!locked)
		spin_lock_irq(&hw->lock);

	ndbg("reply_count %d", dev->reply_count);
        if (!timeout && !dev->read.done) {
		err("orinoco_usb: reply_count %d (resetting)",
		    dev->reply_count);
		dev->reply_count = 0;
                if (dev->write.urb->status != -EINPROGRESS) {
			/* OUT urb is there, but we didn't get the IN
			 * urb back */
                        err("orinoco_usb: timeout (OUT done):"
			    " out status %d in status %d time left %d",
			    dev->write.urb->status, dev->read.urb->status,
			    timeout);
                        retval = -ETIMEDOUT;
                } else {
			err("orinoco_usb: timeout(OUT INPROGRESS, unlinking)");
			spin_unlock_irqrestore(&hw->lock, flags);
			usb_unlink_urb(dev->write.urb);
			spin_lock_irq(&hw->lock);
                        retval = -ETIMEDOUT;
                }
		goto exit;
        } else
                retval = dev->write.urb->status;

	ans = (struct bridge_ans_packet *)dev->read.buff[dev->read.avail_buff];
	dbg("ans->reply_count %d", ans->reply_count);

	if(ans->reply_count != bridge_reply_inc(dev->reply_count)) {
		err("orinoco_usb: incorrect reply count %d/%d",
		    ans->reply_count, bridge_reply_inc(dev->reply_count));
		dev->reply_count=0;
		retval = -EFAULT;
		goto exit;
	}

	dev->reply_count=ans->reply_count;
	ndump((__u8 *)ans, dev->read.avail_actual_length);
	if(((ans->hermes_len+1)*2+10) != dev->read.avail_actual_length)
		err("orinoco_usb: hermes_len=%d actual_length=%d",
		    ans->hermes_len, dev->read.avail_actual_length);
	if(ans_buff)
		memcpy(ans_buff, ans->data,
		       MIN(dev->read.avail_actual_length-sizeof(*ans),
			   ans_size));
	if(ans_length)
		*ans_length = ans->hermes_len;
exit:
	if (!locked)
		spin_unlock_irqrestore(&hw->lock, flags);
	return retval;
}

int bridge_write_ltv(hermes_t *hw, int bap, u16 rid,
		     u16 length, const void *data) 
{
	struct bridge_priv *dev = hw->priv;
	u16 x04 = 0x02;
	
	if (length == 0)
		return -EINVAL;
	ndbg("length: %d", length);
	length = HERMES_RECLEN_TO_BYTES(length);
	ndbg("length: %d", length);
	if (length == 0){
		/* On memory mapped devices
		 * HERMES_RID_CNFGROUPADDRESSES can be set to be
		 * empty, but the USB bridge doesn't like it */
		dbg("zero size data, make them think that we did it");
		return 0;
	}
	
	ndbg("0x%04X", rid);
	if (rid == BRIDGE_RID_TX)
		x04 = 0x01;
	return bridge_access_ltv(dev, length, rid, data, x04, BRIDGE_RID_ACK,
				 NULL, 0, NULL, 0);
}

static int bridge_read_ltv(hermes_t *hw, int bap, u16 rid, unsigned bufsize,
			   u16 *length, void *buf)
{
	struct bridge_priv *dev = hw->priv;
	ndbg("0x%04X", rid);

	if ( (bufsize < 0) || (bufsize % 2) )
		return -EINVAL;

	return bridge_access_ltv(dev, 0, rid, NULL, 0x02, rid,
				 buf, bufsize, length, 0);
}
static int bridge_docmd_wait(hermes_t *hw, u16 cmd, u16 parm0,
			     hermes_response_t * resp, int locked)
{
	struct bridge_priv *dev = hw->priv;
	
	struct {
		u16 cmd;
		u16 parm0;
		u16 parm1;
		u16 parm2;
	} data = {
		.cmd = cmd,
		.parm0 = parm0,
		.parm1 = 0,
		.parm2 = 0,
	};
	dbg("0x%04X, parm0 0x%04X", cmd, parm0);
	ndump((void *)&data, sizeof(data));
	if (cmd == (HERMES_CMD_TX | HERMES_CMD_RECL)){
		int tx_size = dev->write.tx_size;
		dev->write.tx_size = 0;
		ndbg("tx_size: %d", tx_size);
		return bridge_access_ltv(dev, tx_size, BRIDGE_RID_TX,
					 NULL, 0x01, 0, NULL, 0, NULL, locked);
	} else {
		return bridge_access_ltv(dev, sizeof(data), BRIDGE_RID_DOCMD,
					 &data,
					 0x02, BRIDGE_RID_ACK, NULL, 0, NULL,
					 locked);
	}
}
static char *reg_name(u16 off)
{
	switch (off){
	case HERMES_INTEN:
		return "HERMES_INTEN";
	case HERMES_EVSTAT:
		return "HERMES_EVSTAT";
	case HERMES_EVACK:
		return "HERMES_EVACK";
	case HERMES_TXCOMPLFID:
		return "HERMES_TXCOMPLFID";
	case HERMES_ALLOCFID:
		return "HERMES_ALLOCFID";
	case HERMES_SWSUPPORT0:
		return "HERMES_SWSUPPORT0";
	case HERMES_INFOFID:
		return "HERMES_INFOFID";
	default:
		return "";
	}
}
static char *rid_name(u16 rid)
{
	switch (rid){
	case HERMES_INQ_LINKSTATUS:
		return "HERMES_INQ_LINKSTATUS";
	case HERMES_INQ_TALLIES:
		return "HERMES_INQ_TALLIES";
	case HERMES_INQ_SCAN:
		return "HERMES_INQ_SCAN";
	case BRIDGE_RID_RX:
		return "BRIDGE_RID_RX";
	case BRIDGE_RID_TX:
		return "BRIDGE_RID_TX";
	default:
		return "";
	}
}
static u16 bridge_read_reg(struct hermes *hw, u16 off)
{
	struct bridge_priv *dev = hw->priv;
	u16 aux;
	switch (off){
	case HERMES_RXFID:
		aux = BRIDGE_RID_RX;
		break;
	default:
		dbg("off: 0x%04X(%s)", off, reg_name(off));
	case HERMES_EVSTAT:
	case HERMES_SWSUPPORT0:
	case HERMES_ALLOCFID:
	case HERMES_INFOFID:
		aux = dev->hermes_reg_fake[off];
	}
	ndbg("off: 0x%04X(%s) val: 0x%04X", off, reg_name(off), aux);
	return aux;
}
static void bridge_write_reg(struct hermes *hw, u16 off, u16 val)
{
	struct bridge_priv *dev = hw->priv;
	u16 aux;
	ndbg("off: 0x%04X(%s) val: 0x%04X", off, reg_name(off), val);
	switch (off){
	case HERMES_EVACK:
		aux = bridge_read_reg(hw, HERMES_EVSTAT);
		aux &= ~val;
		bridge_write_reg(hw, HERMES_EVSTAT, aux);
		break;
	default:
		err("orinoco_usb: unknown offset: off: 0x%04X(%s) val: 0x%04X",
		    off, reg_name(off), val);
	case HERMES_EVSTAT:
	case HERMES_INTEN:
	case HERMES_SWSUPPORT0:
	case HERMES_ALLOCFID:
	case HERMES_TXCOMPLFID:
	case HERMES_INFOFID:
		dev->hermes_reg_fake[off] = val;
	}
	return;
}
static void bridge_or_reg(struct hermes *hw, u16 off, u16 bits)
{
	u16 aux = bridge_read_reg(hw, off);
	aux |= bits;
	bridge_write_reg(hw, off, aux);
}
static int bridge_bap_pread(struct hermes *hw, int bap,
			    void *buf, unsigned len,
			    u16 id, s32 offset)
{
	struct bridge_priv *dev = hw->priv;
	struct bridge_ans_packet *ans =
		(void *)dev->read.urb->transfer_buffer;
	int actual_length = dev->read.urb->actual_length;
	if(offset < 0){
		dbg("faking seek skip");
		offset = dev->read.bap.offset;
	}
	dbg("bap: %d, id: 0x%04X(%s), offset: 0x%04X, len: %d, size: %d",
	    bap, id, rid_name(id), offset, len, actual_length);
	if (id != BRIDGE_RID_RX){
		ndbg("id == %d != BRIDGE_RID_RX", id);
	}
	if(offset == 0){
		ndump((void *)ans, actual_length);
	}
	switch(id){
	case HERMES_INQ_LINKSTATUS:
	case HERMES_INQ_TALLIES:
	case HERMES_INQ_SCAN:
		/* FIXME: this should be the default */
		if ((sizeof(*ans)+offset+len-4) > actual_length){
			err("orinoco_usb: bap read beyond end actual_length");
			return -EFAULT;
		}
		/* Should include the LENGTH/TYPE part */
		memcpy(buf, (ans->data-4)+offset, len);
		dev->read.bap.offset = offset + len;
		ndbg("dev->read.bap.offset: %d", dev->read.bap.offset);
		break;
	default:
		if ((sizeof(*ans)+offset+len) > actual_length){
			err("orinoco_usb: bap read beyond end actual_length");
			return -EFAULT;
		}
		memcpy(buf, ans->data+offset, len);
	}
	ndump(buf, len);
	return 0;
}

static int bridge_bap_pwrite(struct hermes *hw, int bap, const void *buf,
			     unsigned len, u16 id, u16 offset)
{
	struct bridge_priv *dev = hw->priv;
	struct bridge_req_packet *req = (void *)dev->write.buff;

	dbg("bap: %d, id: 0x%04X, offset: %d, len: %d",
	    bap, id, offset, len);
	if (id != BRIDGE_RID_TX){
		dbg("id != BRIDGE_RID_TX");
		return -EINVAL;
	}
	if(offset == 0){
		memset(dev->write.buff, 0, dev->write.buff_size);
		ndbg("clearing bap");
	}
	if (offset < HERMES_802_3_OFFSET){
		ndbg("skiping some headers");
		return 0;
	}
	offset = offset - HERMES_802_3_OFFSET + 2;
	memcpy(req->data+offset, buf, len);
	dev->write.tx_size = MAX(sizeof(*req)+offset+len, dev->write.tx_size);
	return 0;
}
static void bridge_read_words(struct hermes *hw, int off, void *buf,
			      unsigned count)
{
	struct bridge_priv *dev = hw->priv;
	struct bridge_ans_packet *ans =
		(void *)dev->read.urb->transfer_buffer;
	int actual_length = dev->read.urb->actual_length;
	int offset = dev->read.bap.offset;

	err("orinoco_usb: ERROR read_words SHOULD NOT BE CALLED");
	return;

	dbg("offset: %d, count: %d", off, count);
	count *=2;
	if((sizeof(*ans)+offset+count) > actual_length){
		err("orinoco_usb: bap read beyond end actual_length");
		return;
	}
	memcpy(buf, ans->data+offset, count);
	dev->read.bap.offset += count;
	return;
}
static int bridge_allocate(struct hermes *hw, u16 size, u16 *fid)
{
	struct bridge_priv *dev = hw->priv;
	struct net_device *net_dev = dev->net_dev;

	ndbg("size: %d", size);
	*fid = BRIDGE_RID_TX;
	hermes_write_regn(hw, ALLOCFID, BRIDGE_RID_TX);
	bridge_or_reg(hw, HERMES_EVSTAT, HERMES_EV_ALLOC);
	orinoco_interrupt(net_dev->irq, net_dev, NULL);

	return 0;
}

static int bridge_init(hermes_t *hw)
{
	struct bridge_priv *dev = hw->priv;
	int retval = 0;
	dbg("ENTER");
	if(in_interrupt()){
		err("bridge_init called from interrupt");
		retval = -EFAULT;
		goto exit;
	}
	if(!dev){
		err("NULL dev");
		retval = -EFAULT;
		goto exit;
	}
#if 0
	/* reseting here seams to hand */
	dbg("calling usb_reset_device");
	retval = usb_reset_device(dev->udev);
	if (retval)
		warn(__FUNCTION__ ": usb_reset_device: ret: %d", retval);
	dbg("called usb_reset_device");
#endif
	dev->reply_count=0;
	/* Write the MAGIC number on the simulated registers to keep
	 * orinoco.c happy */
	hermes_write_regn(hw, SWSUPPORT0, HERMES_MAGIC);
	retval = bridge_reset(dev);
	if (retval < 0)
		goto exit;
	dbg("sending control message");
	retval = usb_control_msg(dev->udev,
				 usb_sndctrlpipe(dev->udev, 0),
				 EZUSB_REQUEST_TRIGER,
				 USB_TYPE_VENDOR|USB_RECIP_DEVICE|USB_DIR_OUT,
				 0x0, 0x0,
				 NULL, 0,
				 DEF_TIMEOUT);
	if (retval < 0){
		err("EZUSB_REQUEST_TRIGER failed retval %d", retval);
		goto exit;
	}
	if (dev->flags && ORINOCO_USB_NEEDS_AC_REQ){
		dbg("Sending EZUSB_REQUEST_TRIG_AC");
		retval = usb_control_msg(dev->udev,
					 usb_sndctrlpipe(dev->udev, 0),
					 EZUSB_REQUEST_TRIG_AC,
					 USB_TYPE_VENDOR|USB_RECIP_DEVICE
					 |USB_DIR_OUT,
					 0x00FA, 0x0,
					 NULL, 0,
					 DEF_TIMEOUT);
		if (retval < 0){
			err("EZUSB_REQUEST_TRIG_AC failed retval %d", retval);
			goto exit;
		}
	}
	bridge_remove_in_urb(dev);
	bridge_submit_in_urb(dev);

	retval = bridge_write_ltv(hw, 0, BRIDGE_RID_INIT1,
				  HERMES_BYTES_TO_RECLEN(2), "\x10\x00");
	if (retval < 0){
		err("orinoco_usb: BRIDGE_RID_INIT1 retval %d", retval);
		goto exit;
	}
	retval = hermes_docmd_wait(hw, HERMES_CMD_INIT, 0, NULL);
	if (retval < 0){
		err("orinoco_usb: HERMES_CMD_INIT retval %d", retval);
		goto exit;
	}

exit:
	if(retval < 0){
		if (bridge_reset(dev) < 0){
			err("orinoco_usb: ERROR RESETTING DEVICE");
		}
	}
	dbg("EXIT");
	return retval;
}

static int bridge_netdev_stop(struct net_device *net_dev)
{
	struct orinoco_private *priv = net_dev->priv;
	struct bridge_priv *dev = priv->card;

	if(watchdog_secs)
		del_timer(&dev->watchdog_timer);
	return dev->orinoco_stop(net_dev);
}
static int bridge_netdev_open(struct net_device *net_dev)
{
	struct orinoco_private *priv = net_dev->priv;
	struct bridge_priv *dev = priv->card;
	int retval;

	retval = dev->orinoco_open(net_dev);
	if(watchdog_secs)
		mod_timer(&dev->watchdog_timer, jiffies+watchdog_secs*HZ);
	return retval;
}
/**
 *	usb_skel_debug_data
 */
static inline void usb_skel_debug_data (const char *function, int size,
					const unsigned char *data)
{
	int i;

	if (!debug)
		return;
	
	printk (KERN_DEBUG __FILE__": %s - length = %d, data = ", 
		function, size);
	for (i = 0; i < size; ++i) {
		printk ("%.2x ", data[i]);
	}
	printk ("\n");
}


/**
 *	skel_delete
 */
static inline void skel_delete (struct bridge_priv *dev)
{
	//struct net_device *net_dev = dev->net_dev;
	//struct orinoco_private *priv = net_dev->priv;
	int i;
	
	dbg("ENTER");
	if(in_interrupt()){
		err("orinoco_usb: %s: In interrupt", __FUNCTION__);
		return;
	}
	if(!dev){
		err(__FUNCTION__ ": dev == NULL");
		return;
	}
	del_timer_sync(&dev->watchdog_timer);
	if (dev->net_dev)
		unregister_netdev(dev->net_dev);
	for (i=0; i<2; i++)
		if (dev->read.buff[i] != NULL)
			kfree (dev->read.buff[i]);
	if (dev->write.buff != NULL)
		kfree (dev->write.buff);
	if (dev->read.urb != NULL)
		usb_free_urb(dev->read.urb);
	if (dev->write.urb != NULL)
		usb_free_urb (dev->write.urb);
	if (dev->net_dev)
		kfree(dev->net_dev);
	dbg("EXIT");
}

/**
 *	skel_write_bulk_callback
 */
static void bridge_write_bulk_callback (struct urb *urb)
{
	struct bridge_priv *dev = (struct bridge_priv *)urb->context;
	struct bridge_req_packet *req = (void *)dev->write.buff;
	struct net_device *net_dev = dev->net_dev;
	struct orinoco_private *priv = net_dev->priv;
	hermes_t *hw = &priv->hw;

	ndbg("ENTER");

	if (dev->udev == NULL){
		dbg("disconnected");
		if (dev->read.urb->status != -EINPROGRESS){
			err("orinoco_usb: read urb still in progress");
			//skel_delete(dev);
		}
		return;
	}

	if(watchdog_secs)
		mod_timer(&dev->watchdog_timer, jiffies+watchdog_secs*HZ);

	if(req->hermes_rid == BRIDGE_RID_TX){
		ndbg("BRIDGE_RID_TX status:%d", urb->status);

		if(urb->status)
			bridge_or_reg(hw, HERMES_EVSTAT,
				      HERMES_EV_TXEXC);
		else
			bridge_or_reg(hw, HERMES_EVSTAT,
				      HERMES_EV_TX);
		orinoco_interrupt(net_dev->irq, net_dev, NULL);
		netif_wake_queue(net_dev);
		return;
	}

	if(urb->status) {
		err("orinoco_usb: %s: error status %d", __FUNCTION__,
		    urb->status);
		dev->read.done = 1;
	}
	return;
}

static void bridge_read_bulk_callback (struct urb *urb)
{
	struct bridge_priv *dev = (struct bridge_priv *)urb->context;
	struct net_device *net_dev = dev->net_dev;
	struct orinoco_private *priv = net_dev->priv;
	hermes_t *hw = &priv->hw;
	struct bridge_ans_packet *ans = urb->transfer_buffer;

	if (dev->udev == NULL){
		dbg("disconnected");
		if (dev->write.urb->status != -EINPROGRESS){
			err("orinoco_usb: write urb still in progress");
			//skel_delete(dev);
		}
		return;
	}

	if(watchdog_secs)
		mod_timer(&dev->watchdog_timer, jiffies+watchdog_secs*HZ);

	if (urb->transfer_buffer != dev->read.buff[!dev->read.avail_buff]){
		err(__FUNCTION__ ": ERROR transfer_buffer mismatch %p %p %p",
		    urb->transfer_buffer,
		    dev->read.buff[0], dev->read.buff[1]);
		//BUG();
	}

	if (urb->status == -ETIMEDOUT){
		warn(__FUNCTION__ ": urb timed out, resubmiting");
		goto resubmit;
	}
	if (urb->status == -ECONNABORTED){
		warn(__FUNCTION__ ": connection abort, resubmiting urb");
		goto resubmit;
	}
	if ((urb->status == -EILSEQ)
	    || (urb->status == -ENOENT)){
		dbg("status %d, not resubmiting", urb->status);
		goto exit;
	}
	if(urb->status)
		dbg("status: %d length: %d",
		    urb->status, urb->actual_length);
	if (urb->actual_length < sizeof(*ans)){
		err(__FUNCTION__ ": short read, ignoring");
		goto resubmit;
	}
	if (ans->hermes_rid == dev->read.wait_rid){
		ndbg("got the right RID: 0x%04X", ans->hermes_rid);
		dev->read.done = 1;
		goto resubmit;
	}
	switch (ans->hermes_rid) {
	case BRIDGE_RID_ACK:
		err(__FUNCTION__ ": got unexpected ACK");
		break;
	case BRIDGE_RID_RX:
		ndbg("got a data frame");
		bridge_or_reg(hw, HERMES_EVSTAT, HERMES_EV_RX);
		orinoco_interrupt(net_dev->irq, net_dev, NULL);
		break;
	case HERMES_INQ_LINKSTATUS:
	case HERMES_INQ_TALLIES:
	case HERMES_INQ_SCAN:
		dbg("got an INQ frame %s", rid_name(ans->hermes_rid));
#if 1
		ndump(urb->transfer_buffer, urb->actual_length);
		bridge_or_reg(hw, HERMES_EVSTAT, HERMES_EV_INFO);
		hermes_write_regn(hw, INFOFID, ans->hermes_rid);
		orinoco_interrupt(net_dev->irq, net_dev, NULL);
#endif
		break;
	default:
		err(__FUNCTION__ ": got unexpected RID: 0x%04X(%s)",
		    ans->hermes_rid, rid_name(ans->hermes_rid));
	}
resubmit:
	//if ((dev->net_dev->flags&IFF_UP) || !dev->read.done){
	if (dev->udev){
		dbg("resubmiting IN urb");
		bridge_submit_in_urb(dev);
	} else {
		dbg("NOT resubmiting IN urb");
	}
exit:
	return;
}
static void watchdog_timer_func(u_long data)
{
	struct bridge_priv *dev = (struct bridge_priv *)data;
	struct net_device *net_dev = dev->net_dev;
	struct orinoco_private *priv = net_dev->priv;
	hermes_t *hw = &priv->hw;

	if (!netif_running(net_dev)){
		/* Some lost timer, of course there is no traffic if the
		 * interface is not running */
		dbg("!netif_running(net_dev): ignoring timer");
		return;
	}
	err("card stuck? sending an INQ command to get it going (ret:%d)",
	    hermes_inquire(hw, HERMES_INQ_TALLIES));

#if 0
	/* The timer get's restarted often enough without this, and resetting
	 * it here, makes it a little too agresive, altough it does work */
	if(watchdog_secs){
		mod_timer(&dev->watchdog_timer, jiffies+watchdog_secs*HZ);
	}
#endif
}

/**
 *	skel_probe
 *
 *	Called by the usb core when a new device is connected that it thinks
 *	this driver might be interested in.
 */
static void * skel_probe(struct usb_device *udev, unsigned int ifnum,
			 const struct usb_device_id *id)
{
	struct orinoco_private *priv;
	struct net_device *net_dev;
	hermes_t *hw;
	struct bridge_priv *dev = NULL;
	struct usb_interface *interface = &udev->actconfig->interface[ifnum];
	struct usb_interface_descriptor *iface_desc;
	struct usb_endpoint_descriptor *ep;
	int buffer_size = BULK_BUF_SIZE;
	int card_variant = -1;
	const struct fwfs_entry *fwfs_entry_new;
	int i;

	/* See if the device offered us matches what we can accept */
	
	if (usb_match_id(udev, interface, bridge_varian1_table))
		card_variant = 0;
	if (usb_match_id(udev, interface, bridge_varian2_table))
		card_variant = 1;

	if(force_unsupported)
		card_variant = -1;

	if (card_variant == -1) {
		err("Device is not supported (you may want to set"
		    " force_unsupported=1)");
		if(!force_unsupported)
			return NULL;
		err("Trying to handle device anyway as requested");
	}

	dbg("ENTER");
	MOD_INC_USE_COUNT;
	down (&minor_table_mutex);

	if((fwfs_entry_new = fwfs_get("orinoco_ezusb_fw"))){
		fwfs_put(fwfs_entry);
		if (fwfs_entry != fwfs_entry_new){
			fwfs_entry = fwfs_entry_new;
			firmware.size = fwfs_entry->size;
			firmware.code = fwfs_entry->data;
			find_fw_variant_offset(&firmware);
		}
	}

	net_dev = alloc_orinocodev(sizeof(*dev), NULL);
	if (!net_dev){
		err("Couldn't allocate net_dev");
		goto exit;
	}
	
	
	priv = net_dev->priv;
	hw = &priv->hw;
	
	dev = priv->card;
	dev->net_dev = net_dev;


	init_MUTEX (&dev->sem);
	dev->watchdog_timer.function = watchdog_timer_func;
	dev->watchdog_timer.data = (u_long)dev;
	init_timer(&dev->watchdog_timer);

	dev->udev = udev;
	dev->interface = interface;

	hermes_struct_init(hw, (ulong)&dev->hermes_reg_fake,
			   HERMES_MEM, HERMES_16BIT_REGSPACING);
	hw->priv = dev;

	hw->read_reg = bridge_read_reg;
	hw->write_reg = bridge_write_reg;
	hw->init = bridge_init;
	hw->docmd_wait = bridge_docmd_wait;
	hw->allocate = bridge_allocate;
	hw->read_ltv = bridge_read_ltv;
	hw->write_ltv = bridge_write_ltv;
	hw->bap_pread = bridge_bap_pread;
	hw->bap_pwrite = bridge_bap_pwrite;
	hw->read_words = bridge_read_words;
	hw->write_words = NULL;
	//net_dev->init = orinoco_init;
	dev->orinoco_stop = net_dev->stop;
	net_dev->stop = bridge_netdev_stop;
	dev->orinoco_open = net_dev->open;
	net_dev->open = bridge_netdev_open;
	//net_dev->hard_start_xmit = bridge_xmit;
	//net_dev->set_multicast_list = NULL;

#if 0
	net_dev->base_addr = (ulong)&dev->hermes_reg_fake;
	net_dev->irq = -1;
#endif
	SET_MODULE_OWNER(net_dev);
	strcpy(net_dev->name, "wlan%d");

	/* set up the endpoint information */
	/* check out the endpoints */
	iface_desc = &interface->altsetting[0];
	for (i = 0; i < iface_desc->bNumEndpoints; ++i) {
		ep = &iface_desc->endpoint[i];

		if (((ep->bEndpointAddress & USB_ENDPOINT_DIR_MASK)
				== USB_DIR_IN) &&
		    ((ep->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
		     		== USB_ENDPOINT_XFER_BULK)) {
			int i;
			/* we found a bulk in endpoint */
			if (dev->read.urb != NULL) {
				warn("Found a second bulk in ep, ignored");
				continue;
			}

			dev->read.done = 1;

			dev->read.urb = usb_alloc_urb(0);
			if (!dev->read.urb) {
				err("No free urbs available");
				goto error;
			}
			if (ep->wMaxPacketSize != 64)
				warn("bulk in: wMaxPacketSize!= 64");
			dev->read.buff_size = buffer_size;
			dev->read.pipe = ep->bEndpointAddress;
			if (dev->read.pipe != (2|USB_DIR_IN))
				warn("bulk in: bEndpointAddress: %d",
						dev->read.pipe);
			for (i=0; i<2; i++) {
				dev->read.buff[i] = kmalloc (buffer_size,
							     GFP_KERNEL);
				if (!dev->read.buff[i]) {
					err("Couldn't allocate IN buffer");
					goto error;
				}
			}
			usb_fill_bulk_urb(
				dev->read.urb, udev, 
				usb_sndbulkpipe(udev, ep->bEndpointAddress),
				dev->read.buff[0], buffer_size,
				bridge_read_bulk_callback, dev);
		}
		
		if (((ep->bEndpointAddress & USB_ENDPOINT_DIR_MASK)
				== USB_DIR_OUT) &&
		    ((ep->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
		     		== USB_ENDPOINT_XFER_BULK)) {
			/* we found a bulk out endpoint */
			if (dev->write.buff != NULL) {
				warn("Found a second bulk out ep, ignored");
				continue;
			}

			dev->write.urb = usb_alloc_urb(0);
			if (!dev->write.urb) {
				err("No free urbs available");
				goto error;
			}
			if (ep->wMaxPacketSize != 64)
				warn("bulk out: wMaxPacketSize != 64");
			dev->write.buff_size = buffer_size;
			dev->write.pipe = ep->bEndpointAddress;
			if (dev->write.pipe != 2)
				warn("bulk in: bEndpointAddress: %d",
						dev->read.pipe);
			dev->write.buff = kmalloc (buffer_size,
							GFP_KERNEL);
			if (!dev->write.buff) {
				err("Couldn't allocate bulk_out_buffer");
				goto error;
			}
			usb_fill_bulk_urb(
				dev->write.urb, udev, 
				usb_sndbulkpipe(udev, ep->bEndpointAddress),
				dev->write.buff, buffer_size,
				bridge_write_bulk_callback, dev);
			dev->write.urb->timeout = DEF_TIMEOUT;
		}
	}
	if (!dev->write.urb
	    || !dev->read.urb){
		err("Didn't find the required bulk endpoints");
		goto error;
	}

	if(card_variant != -1)
		dev->card_variant = card_variant;
	else
		dev->card_variant = default_card_variant;

	/* I plan on making in-kernel firmware download optional, so
	 * the souces don't have to go with the firmware */
	if(firmware.size)
		bridge_firmware_download(dev, &firmware);
	else {
		err("No firmware to download");
		goto error;
	}

	/* If the firmware is already downloaded orinoco.c will call
	 * bridge_init but if the firmware is not already there, that will make
	 * the kernel very unstable, so we try initializing here and quit in
	 * case of error */
	if (bridge_init(hw) < 0){
		err("Couldn't initialize the device");
		err("Firmware may not be downloaded or may be wrong.");
		if(card_variant == -1){
			err("Replug the device to pretend it is the other"
			    " variant");
			default_card_variant = !default_card_variant;
		}
		bridge_remove_in_urb(dev);
		goto error;
	}
	if (card_variant == -1){
		err("Your device variant seams to be 0x%02X, please send this"
		    " message and the output of 'lsusb -v' to"
		    " orinoco-usb-devel@ranty.ddts.net", dev->card_variant);
	}

	dbg("registering net_dev");
	if (register_netdev(net_dev) != 0) {
		err(__FUNCTION__ ": register_netdev() failed");
		goto error;
	}
	dbg("net_dev registered");

	/* let the user know what node this device is now attached to */
	info ("USB Skeleton device now attached");
	goto exit;
	
error:
	skel_delete (dev);
	dev = NULL;
	MOD_DEC_USE_COUNT;
exit:
	up (&minor_table_mutex);
	dbg("EXIT");
	return dev;
}


/**
 *	skel_disconnect
 *
 *	Called by the usb core when the device is removed from the system.
 */
static void skel_disconnect(struct usb_device *udev, void *ptr)
{
	struct bridge_priv *dev = ptr;
	struct net_device *net_dev = NULL;
	struct orinoco_private *priv = NULL;
	int retval;

	if(!dev){
		dbg("NULL device pointer");
		return;
	}
	net_dev = dev->net_dev;
	priv = net_dev->priv;
	down (&minor_table_mutex);
	down (&dev->sem);


	if(in_interrupt()){
		dbg("in_interrupt");
	}
	//priv->hw_unavailable = 1;

	if(dev->read.urb->status == -EINPROGRESS){
		warn(__FUNCTION__ ": Unlinking pending IN urb");
		retval = bridge_remove_in_urb(dev);
		if(retval){
			dbg("retval %d status %d", retval,
			    dev->read.urb->status);
		}
	}
	if(dev->write.urb->status == -EINPROGRESS){
		warn(__FUNCTION__ ": Unlinking pending OUT urb");
		retval = usb_unlink_urb(dev->write.urb);
		if(retval){
			dbg("retval %d status %d", retval,
			    dev->write.urb->status);
		}
	}
#if 0
	/* I think that reseting here is also giving trouble */
	retval = usb_reset_device(udev);
	if (retval)
		dbg("usb_reset_device: ret: %d", retval);
#endif
	if ((dev->read.urb->status == -EINPROGRESS)
	    || (dev->write.urb->status == -EINPROGRESS)) {
		err("orinoco_usb: Some URB in progress");
	}
	dev->udev = NULL;
	up (&dev->sem);
	skel_delete (dev);

	info("USB Skeleton now disconnected");
	up (&minor_table_mutex);
	MOD_DEC_USE_COUNT;
}
/**
 *	usb_skel_init
 */

static int __init usb_skel_init(void)
{
	int result;

	//THIS_MODULE->can_unload = bridge_can_unload;
	if(firmware.size)
		fwfs_write_default("orinoco_ezusb_fw", firmware.code,
				   firmware.size);
	/* register this driver with the USB subsystem */
	result = usb_register(&skel_driver);
	if (result < 0) {
		err("usb_register failed for the "
		    __FILE__" driver. Error number %d",
		    result);
		return -1;
	}

	info(DRIVER_DESC " " DRIVER_VERSION);
	return 0;
}


/**
 *	usb_skel_exit
 */
static void __exit usb_skel_exit(void)
{
	fwfs_put(fwfs_entry);
	/* deregister this driver with the USB subsystem */
	usb_deregister(&skel_driver);
}


module_init (usb_skel_init);
module_exit (usb_skel_exit);

MODULE_AUTHOR("Manuel Estrada Sainz <ranty@debian.org>");
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");

